/*************************************************************************

	Project : ExpandMe
	Author  : R'alf <mailto:raphael.moll@inforoute.cgs.fr>

	Version : 0.6.5
	Date    : 01 Septembre 1998

	Format  : tabs == 2	

	Description :
	- An "expand" add-on for AA-DR9 Tracker.
	- It applies "tar xf" on .tar files, gunzip on .tar?gz and .gz files,
	  and "unzip" on .zip files.
	- Recurse througth subdirectories.
	- Expand all selected files.
	- Expand a whole folder if one is selected.
	- Now has some GUI (with as little chrome as possible) for status.

	Version 0.6.5 :
	- new main() installer (moves the executable in user addon dir)

	Version 0.6.1 :
	- added support for bzip/bunzip for files ".bz".
	- bzip is one BeOS thanks to the effort of Michael Griebling
	  (aka grieblm@inforamp.net). You will find it somewhere in BeWare.
	  You must have a "bunzip" binary somewhere in your path.

	Version 0.6 :
	- now gracefully handles files containing spaces, quotes, double
	  quotes and others characters interpreted by the shell. Thx to GL
	  and Ludo !

	For example, the file :
	"./#zi|;p ~m^ &$e_06 _P&R.t)xt.z ip.z_ @ip.z\}ip.zip&(-_.zip)=~#{[|`\^@]}^$¨$*%!;,?.§<stuff>dyn.zip"
	can now be correctly uzipped (I've tried ;))

	Based on code from TermHire.cpp, v0.97, release date 24/05/97 by
	Pierre BRUA (brua@dess-info.u-strasbg.fr

	USE AT YOUR OWN RISK !

*************************************************************************/

#include <Directory.h>
#include <Window.h>
#include <StringView.h>
#include <File.h>
#include <Alert.h>
#include <Message.h>
#include <Errors.h>
#include <Path.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

//------------------------------------------------------------------------
// some defines...

#pragma once

// use this define if you want little windows to explain you what's going on
#define VERBOSE

// use this define if you want printf() in the terminal (run a terminal,
// use Olivier's TManager to quit the Tracker, type /boot/system/Tracker & in
// the terminal and you get the stdout here. Too cool !
#undef DPRINTF

// the name of the add_on
#define ADDON_NAME "ExpandMe"
#define ADDON_VERS "0.6.5"


#ifndef max
#define max(a,b) ((a)>(b) ? (a) : (b))
#endif

//------------------------------------------------------------------------
// rules list
struct a_rule
{
	char *ext;
	char *command;
};


//------------------------------------------------------------------------
// internal state. I use a struct rather than global variables.

struct SState
{
	BWindow *win;
	BTextControl *leaf, *dir, *method, *number;
	BStatusBar *bar;
	BTextView  *console;
	int32 errorCount;
	int32 totalCount;
	int32 baseCount;
	BList	ruleList;		// a list of a_rule ptrs
};

//------------------------------------------------------------------------
// exported function -- the one required by a Tracker Add-on

// this tell the compiler that process_refs will be called from
// outside (by the Tracker)

#include <TrackerAddOn.h>


//------------------------------------------------------------------------
// local prototypes 
int expandFile(const char *leaf, const char *dir, SState &state);
int expandDirectory(SState &state, BDirectory &dir);
void error(const char *message);
void createWindow(SState &state);
void closeWindow(SState &state);
void setFileInfo(SState &state, const char *leaf, const char *dir);
void setError(SState &state, const char *error);
void setMethod(SState &state, const char *text);
void setNumber(SState &state, int32 count);
void printConsole(SState &state, const char *string);
void installOurself(void);


//********************
int main(int, char **)
//********************
/* Code to help a little the fellow user double-clicking the
   add-on icon directly.
*/
{
	new BApplication("application/x-vnd.ralf-vanilla");
	BAlert *box = new BAlert(ADDON_NAME,
													 "Thanks for using " ADDON_NAME " " ADDON_VERS " !\n\n"
													 "Please move the " ADDON_NAME " executable into\n"
													 "/boot/home/config/add-ons/Tracker/\n"
													 "and try again !",
													 "Don't Touch Me !", "Install Myself !", NULL,
													 B_WIDTH_FROM_WIDEST);
	if (box->Go() == 1) installOurself();

	delete be_app;
	return 0;
}


//***********************
void installOurself(void)
//***********************
/*
	Copy *this* binary file into /boot/home/config/add-ons/Tracker/,
	making sure the directories exists and can be created.
*/
{
char appName[B_FILE_NAME_LENGTH];
app_info info_app;
BEntry file_app;
BPath	idealPath;
BDirectory addonsDir;
BDirectory trackerDir;
status_t result;

	be_app->GetAppInfo(&info_app);
	result=file_app.SetTo(&info_app.ref);
	if (result != B_NO_ERROR) goto bail_out_error;

	// The ideal directory is B_USER_ADDONS_DIRECTORY+"./Tracker"
	// create it if doesn't exists

	result = find_directory(B_USER_ADDONS_DIRECTORY, &idealPath, true);
	if (result != B_NO_ERROR && result != B_FILE_EXISTS) goto bail_out_error;

	if (addonsDir.SetTo(idealPath.Path()) != B_NO_ERROR) goto bail_out_error;
	result = addonsDir.CreateDirectory("Tracker", &trackerDir);
	if (result == B_FILE_EXISTS) result = trackerDir.SetTo(&addonsDir, "Tracker");
	if (result != B_NO_ERROR) goto bail_out_error;
	if (trackerDir.InitCheck() != B_NO_ERROR) goto bail_out_error;

	// move executable
	result = file_app.MoveTo(&trackerDir, NULL, true);
	if (result != B_NO_ERROR && result != B_FILE_EXISTS) goto bail_out_error;

	return;

bail_out_error:
	error("Couldn't toss the directories and the file as expected.");
}


//*****************************
void error(const char *message)
//*****************************
/*
	Displays a message in a dialog box (yes, believe it!)
*/
{
	BAlert *alert = new BAlert(ADDON_NAME " Error :\n", message, "OK");
	alert->Go();
}


//********************************************************************************
int expandFile(const char *leaf, const char *dir, SState &state)
//********************************************************************************
{
char buf[1024];
char buf2[1024];
char tmpbuf[1024];
long len;

	setFileInfo(state, leaf, dir);

	// chdir might not be necessary -- unless I do recursive expand
	chdir(dir);

	if(strchr(leaf,'\'') != NULL)
		sprintf(buf2, "\"%s\"", leaf);
	else
		sprintf(buf2, "'%s'", leaf);

	// a macro that checks the postfix : 'n' is strlen('x')
	len = strlen(leaf);
	#define POSTFIX(x,n) (len > n && strncasecmp(&leaf[len-n], x, n) == 0) // or use strncmp

	// create a temporary file name
	// Note that tmpnam() creates a file name in the form "/tmp/tempXXXXXXXXX"
	// thus I just need to add "/var" to get the file in /var/tmp :-)
	sprintf(tmpbuf, "/var%s", tmpnam(NULL));

	if (POSTFIX(".tar.gz", 7) || POSTFIX(".tgz", 4))
	{
		sprintf(buf, "zcat %s | tar xf -", buf2);
		setMethod(state, "zcat -> untar");
	}
	else if (POSTFIX(".tar", 4))
	{
		sprintf(buf, "tar xf %s", buf2);
		setMethod(state, "untar");
	}
	else if (POSTFIX(".gz", 3))
	{
		sprintf(buf, "gunzip %s", buf2);
		setMethod(state, "gunzip");
	}
	else if (POSTFIX(".zip", 4))
	{
		sprintf(buf, "unzip -o %s", buf2);
		setMethod(state, "unzip");
	}
	else if (POSTFIX(".bz", 3))
	{
		sprintf(buf, "bunzip %s", buf2);
		setMethod(state, "bunzip");
	}
	else if (POSTFIX(".uue", 4))
	{
		sprintf(buf, "uudecode %s", buf2);
		setMethod(state, "uudecode");
	}
	else if (POSTFIX(".lha", 4))
	{
		sprintf(buf, "lha x -f %s", buf2);
		setMethod(state, "lha");
	}
	else if (POSTFIX(".jar", 4))
	{
		sprintf(buf, "jar x %s", buf2);
		setMethod(state, "jar");
	}
	else if (POSTFIX(".arj", 4))
	{
		sprintf(buf, "arj x %s", buf2);
		setMethod(state, "arj");
	}
	//
	// Peter Hinely <phinely@hawaii.edu> 14 Jul 1998
	//
	else if (POSTFIX(".bz2", 4)) 
	{
		sprintf(buf, "bunzip2 %s", buf2); 
		setMethod(state, "bunzip2"); 
	}
	else if (POSTFIX(".boz", 4)) 
	{
		sprintf(buf, "bunzip2 < %s > %s;\n"
								 "unzip -o %s;\nrm %s", buf2, tmpbuf, tmpbuf, tmpbuf);
		setMethod(state, "boz"); 
	} 
	else
	{
		printConsole(state, "Unknow file type for ");
		printConsole(state, buf2);
		printConsole(state, "\n");
		setError(state, "Can't handle file - wrong type");
		state.errorCount++;
		return B_NO_ERROR;
	}

	// issue the "system(buf)"
	
	if (state.console)
	{
		printConsole(state, "Executing ");
		printConsole(state, buf);
		printConsole(state, "\n---------\n");

		FILE *f = popen(buf, "rt");
		if(f)
		{
			while(!feof(f))
			{
				fgets(buf2, 512, f);
				printConsole(state, buf2);
			}
			pclose(f);
		}

		printConsole(state, "\n---\nEnd");
	}
	else system(buf);

	return B_NO_ERROR;
}



//********************************************************************************
int expandDirectory(SState &state, BDirectory &dir)
//********************************************************************************
{
BDirectory dir2;
int32 count;
status_t status;
BEntry entry;
BFile file;

	status = dir.InitCheck();
	if (status >= B_NO_ERROR) status = dir.Rewind();
	if (status < B_NO_ERROR) return status;

	count = dir.CountEntries();
	if (count < 1) return B_NO_ERROR;
	state.baseCount = state.totalCount;
	state.totalCount += count;

		#ifdef VERBOSE
		if (state.win && state.bar)
		{
			state.win->Lock();
			state.bar->SetMaxValue(state.totalCount);
			state.win->Unlock();
		}
		#endif

	for (int ix=0; ix<count && dir.GetNextEntry(&entry) >= B_NO_ERROR; ix++)
	{
		#ifdef VERBOSE
			setNumber(state, ix);
		#endif

		// if the ref is a folder, expand the directory content
		if(dir2.SetTo(&entry) >= B_NO_ERROR)
		{
			expandDirectory(state, dir2);
		}
		// if the ref is not a folder, launch a Terminal on the parent folder
		else if(entry.InitCheck() >= B_NO_ERROR) 
		{
			// I ask for the parent directory of the entry
			BEntry entry2;
			BPath path, path2;
			entry.GetParent(&entry2);
			entry.GetPath(&path);
			entry2.GetPath(&path2);
			expandFile(/*path.Path(),*/ path.Leaf(), path2.Path(), state);
		}
	}

	return B_NO_ERROR;
}


//********************************************************
void process_refs(entry_ref dir_ref, BMessage* msg, void*)
//********************************************************
/*
	this is the function the Tracker call when the add-on is selected
	dir_ref is the folder you have launched it from
	msg is a BMessage containing the references to BEntry's(Files/Folders)
	and for "void *", I don't know yet...
*/
{
int32 count;	// number of references stored in the BMessage
ulong type;		// type of the BMessage
SState state;
BPath path;

	// be clean and please respect the OS.h header allegory.
	if (!is_computer_on())
	{
		error("Please power-on your computer before running " ADDON_NAME);
		return;
	}

	// might check is_computer_on_fire() to be clean, too

	state.win = NULL;
	state.errorCount = 0;
	state.totalCount = 0;
	state.baseCount = 0;

	// this creates an informative window about the completion of the add-on
	#ifdef VERBOSE
		createWindow(state);
	#endif

	// first, we consider the case when the add-on has been selected
	// from the file menu of the browser window or by right-clicking
	// in the background of the window
	// i.e : in this case, there is no refs on the BMessage
	msg->GetInfo("refs", &type, &count);
	if (count == 0)
	{
		error("Please select a file or a directory to expand !");
	}
	else
	{
		BDirectory dir;
		BEntry entry;

		// if there are datas on the BMessage, but not from the
		// directory type, exit...
		if (dir.SetTo(&dir_ref) != B_NO_ERROR) 
		{
			error("You must launch " ADDON_NAME " from folders.");
			goto bail_out;
		}
		if(dir.GetEntry(&entry) != B_NO_ERROR)
		{
			error("Sorry, I cannot access the folder while launching " ADDON_NAME);
			goto bail_out;
		}

		state.totalCount = count;

		#ifdef VERBOSE
		if (state.win && state.bar)
		{
			state.win->Lock();
			state.bar->SetMaxValue(count);
			state.win->Unlock();
		}
		#endif

		entry_ref ref;
		BFile file;
		for (int ix=0; ix<count; ix++)
		{
			if (msg->FindRef("refs", ix, &ref) < B_NO_ERROR) continue;

			#ifdef VERBOSE
				setNumber(state, ix);
			#endif

			// if the ref is a folder, expand the directory content
			if(dir.SetTo(&ref) >= B_NO_ERROR)
			{
				expandDirectory(state, dir);
			}
			// if the ref is not a folder, launch a Terminal on the parent folder
			else if(entry.SetTo(&ref) >= B_NO_ERROR) 
			{
				// I ask for the parent directory of the entry
				BEntry entry2;
				BPath path2;
				entry.GetParent(&entry2);
				entry.GetPath(&path);
				entry2.GetPath(&path2);
				if(expandFile(/*path.Path(),*/ path.Leaf(), path2.Path(), state) != B_NO_ERROR) break;
			}
		}
	}
	
	// exit...

	if (state.errorCount > 0)
	{
		char s[256];
		sprintf(s, ADDON_NAME " v." ADDON_VERS " :\n%d error%s detected.",
			state.errorCount, (state.errorCount > 1 ? "s" : ""));
		#ifdef VERBOSE
			if (state.win) state.win->Activate();
		#endif
		error(s);
	}

	// this is a hack for debug purposes. I don't think it's a good feature to keep in.
	// In next version I should better derive a class from BWindow and implement QuitRequested.
	if(modifiers() & B_SHIFT_KEY)
	{
		printConsole(state,"\nRelease the SHIFT key to close the window.\n");
		while(modifiers() & B_SHIFT_KEY) snooze(20000);
	}

bail_out:
	#ifdef VERBOSE
		closeWindow(state);
	#endif
}


//******************************
void createWindow(SState &state)
//******************************
{
	BRect area(0,0, 320,80+250);

	BScreen screen;		// -- DR9 PR
	BRect r=screen.Frame();
	area.OffsetBy((r.Width()-area.Width())/2, (r.Height()-area.Height())/2);

	state.win = new BWindow(area, ADDON_NAME " Status", B_TITLED_WINDOW, B_NOT_CLOSABLE);
	if (!state.win) return;

	BRect bounds = state.win->Bounds();
	BView *base = new BView(bounds, B_EMPTY_STRING, B_FOLLOW_ALL, B_WILL_DRAW);
	base->SetViewColor(230, 230, 230);
	if (!base) goto error;
	state.win->AddChild(base);

	BRect rect;
	rect.left		= bounds.left  + 10.0;
	rect.top		= bounds.top   + 10.0;
	rect.right	= bounds.right - 10.0;
	rect.bottom	= rect.top+15;

	uint32 rmask = B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP;
	BStringView *str= new BStringView(rect, B_EMPTY_STRING,
										ADDON_NAME " v." ADDON_VERS ", by R'alf", rmask, B_WILL_DRAW);
	if (!str) goto error;	
	base->AddChild(str);

	rect = str->Frame();
	rect.OffsetBy(0.0, rect.Height()+ 20.0);
	state.leaf		= new BTextControl(rect, B_EMPTY_STRING, "File name", "", NULL, rmask);
	if (!state.leaf) goto error;
	base->AddChild(state.leaf);

	rect = state.leaf->Frame();
	rect.OffsetBy(0.0, rect.Height()+ 10.0);
	state.dir			= new BTextControl(rect, B_EMPTY_STRING, "Directory", "", NULL, rmask);
	if (!state.dir) goto error;
	base->AddChild(state.dir);

	rect.OffsetBy(0.0, rect.Height()+ 10.0);
	state.method	= new BTextControl(rect, B_EMPTY_STRING, "Running", "", NULL, rmask);
	if (!state.method) goto error;
	base->AddChild(state.method);

	state.number = NULL;

	rect.OffsetBy(0.0, rect.Height()+10.0);
	state.bar = new BStatusBar(rect, B_EMPTY_STRING, "File count", NULL /*, B_WILL_DRAW, rmask*/);
	if (!state.bar) goto error;
	state.bar->SetFlags(B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE);
	state.bar->SetResizingMode(rmask);
	base->AddChild(state.bar);
	rect = state.bar->Frame();

	rect.OffsetBy(0.0, rect.Height()+10.0);
	rect.right -= 14;
	rect.bottom = bounds.bottom-10-14;
	rgb_color noir={0,0,0};
	BRect rect2(0,0,r.Width()-60,r.Height());
	state.console = new BTextView(rect, B_EMPTY_STRING, rect2,
									be_fixed_font, &noir,
									B_FOLLOW_ALL, B_WILL_DRAW);
	if (!state.console) goto error;
	BScrollView *scroll = new BScrollView("scroller",
											state.console,
											B_FOLLOW_ALL, B_WILL_DRAW,
											true,true);
	if (!scroll) goto error;
	state.console->MakeEditable(false);
	state.console->MakeSelectable(false);
	base->AddChild(scroll);

	long n=20;
	n = max(state.leaf   ->StringWidth("File name" ), n);
	n = max(state.dir    ->StringWidth("Directory" ), n);
	n = max(state.method ->StringWidth("Running"   ), n);
	state.leaf   ->SetDivider(n + 2.0);
	state.dir    ->SetDivider(n + 2.0);
	state.method ->SetDivider(n + 2.0);

	state.leaf   ->SetFlags(B_WILL_DRAW);
	state.dir    ->SetFlags(B_WILL_DRAW);
	state.method ->SetFlags(B_WILL_DRAW);

	state.dir   ->TextView()->MakeEditable  (false);
	state.dir   ->TextView()->MakeSelectable(false);
	state.method->TextView()->MakeEditable  (false);
	state.method->TextView()->MakeSelectable(false);
	state.leaf  ->TextView()->MakeEditable  (false);
	state.leaf  ->TextView()->MakeSelectable(false);

	state.win->ResizeTo(bounds.Width(), rect.bottom+10.0);
	state.win->SetSizeLimits(bounds.Width(), r.Width()-20.0, rect.bottom+10.0, r.Height()-20.0);
	state.win->Show();
	return;

error:
	state.win = NULL;

} // end of createWindow


//******************************
void 	closeWindow(SState &state)
//******************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;

	state.win->Lock();
	state.win->Quit();
	int32 val;
	wait_for_thread(state.win->Thread(), &val);

} // end of closeWindow



//****************************************************************
void 	setFileInfo(SState &state, const char *leaf, const char *dir)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	if (leaf && state.leaf) state.leaf->SetText(leaf);
	if (dir  && state.dir ) state.dir ->SetText(dir );
	state.win->Unlock();
} // end of setFileInfo


//****************************************************************
void 	setError(SState &state, const char * /*error*/)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	// DO SOMETHING : print in the console
	state.win->Unlock();
} // end of setError


//****************************************************************
void 	setMethod(SState &state, const char *text)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	if (text && state.method) state.method->SetText(text);
	state.win->Unlock();
} // end of setMethod


//****************************************************************
void 	setNumber(SState &state, int32 count)
//****************************************************************
{
char s[256];
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	sprintf(s, "%d of %d", state.baseCount+count, state.totalCount);
	state.win->Lock();
	if (state.bar) state.bar->Update(1.0, NULL, s);
	state.win->Unlock();
} // end of setNumber


//****************************************************************
void 	printConsole(SState &state, const char *string)
//****************************************************************
{
#ifndef VERBOSE
	return;
#endif
	if (!state.win) return;
	state.win->Lock();
	state.console->Insert(string);
	state.console->ScrollToSelection();
	state.win->Unlock();
} // end of printConsole

// eoc

